# -*- coding: utf-8 -*-
'''
Bikini: a minimalist Python wiki.
'''
from __future__ import with_statement # Python 2.5+
import sys, os, re, time, codecs, cgi, string, hashlib, urllib, time, timeformat
from wsgiref.util import application_uri
from wsgiref.headers import Headers
from Cookie import SimpleCookie
from datetime import datetime
from traceback import format_tb
from bikini import storage, config

__author__ = 'Andrea Peltrin'
__version__ = (2, 0, 0, '-dev')
__license__ = 'GPL'

VERSION_STRING = '%d.%d.%d%s' % __version__
ANONYNOUS_USER = 'AnonymousCoward'
RESERVED_NAMES = 'AboutBikini Maintenance Names Needed RecentChanges Unlinked UserPreferences'.split()
EMPTY_HASH = 'd41d8cd98f00b204e9800998ecf8427e'
WIKI_WORD = '[A-Z]+[a-zA-Z0-9]+' 
ENCODING = 'utf-8'
DEBUG = False
REVISION_LIFESPAN_DAYS =  config.getint('revisions', 'revision_lifespan_days')  

SETUP_PATH = './bikini/setup'
TEMPLATE_PATH = ['./bikini/templates'] #  Support for multiple template dirs

SITE_LANG = 'en'
LOG_PATH = config.get('advanced', 'log_path')
DATABASE_PATH = config.get('advanced', 'database_path')

DEBUG = True
OWNER_URI = config.get('owner', 'owner_uri')
OWNER_NAME = config.get('owner', 'owner_name')
SITE_NAME = u'Your Wiki'
SITE_DESCRIPTION = u'Yet another Bikini wiki'
USE_SMARTYPANTS = False  
NAV_NAMES = 'FrontPage Names RecentChanges UserPreferences AboutBikini'



HTTP_OK = '200 OK'
HTTP_NOT_FOUND = '404 Not Found'
HTTP_SEE_OTHER = '303 See Other'
# HTTP_MOVED_PERMANENTLY = '301 Moved Permanently'
HTTP_INTERNAL_SERVER_ERROR = '500 Internal Server Error'
HTTP_FORBIDDEN = '403 Forbidden'
                
re_wiki_word = re.compile(ur'^%s$' % WIKI_WORD) 

def log(environ, value):
    if not DEBUG: return #Silent
        
    from pprint import pprint
    pprint(value, stream=environ['wsgi.errors'])
 
# ------------------------------------------------------
# Response
# ------------------------------------------------------

URI_MAP = []
    
# Decorator
def view(pattern='^$', request_method='GET'):    
    
    def _(handler):         
        URI_MAP.append((re.compile(pattern, re.U), request_method, handler))
        return handler   
    return _   
    
def dispatch_request(environ, start_response):

    path_info = environ.get('PATH_INFO', '')
    
    output_headers = Headers([('Content-Type', 'text/html; charset=%s' % ENCODING)])

    template_symbols = {
        'base_uri': get_base_uri(environ), 
        'meta': meta,
        'header': header,
        'footer': footer,
        'ENCODING': ENCODING,
        'VERSION_STRING': VERSION_STRING,
        'SITE_LANG': SITE_LANG,
        'SITE_NAME': config.get('wiki', 'wiki_name'),
        'SITE_DESCRIPTION': config.get('wiki', 'wiki_description'),
        'OWNER_NAME': config.get('owner', 'owner_name'),
        'OWNER_URI': config.get('owner', 'owner_uri'),
    }
    
    # Setup common template symbols
    filler = fill_template_namespace(**template_symbols)
    
    def find_view():
    
        request_method = environ['REQUEST_METHOD']
        
        for re, method, handler in URI_MAP:        
            match = re.match(path_info)
            if match and (method == request_method):                    
                return handler, match.groups()
    
        raise NotFoundError  
    
    def response(lines):
        for line in lines:
            yield encode(line)

    try:                    
        handler, args = find_view()
                   
        if not args:
            args = ()

        status, headers, body = handler(environ, filler, *args)

        for name, value in headers:
            if name == 'Content-Type':
                output_headers[name] = value # Replace default value
            else:
                output_headers.add_header(name, value)

    except NotFoundError, exc:     
        status = str(exc)
        if exc.body:
            message = exc.body
        else:
            message = u'Resource <i>%s</i> could not be found on this server.' % path_info
                    
        body = make_page('404.html', filler, {'message': message})

    #output_headers['Content-Length'] = str(len(body))

    #@@NOTE: Looks wrong to me. However, WSGI handlers.py uses StringType 
    #  instead of str while asserting for k and v types
    start_response(status, [(str(k), str(v)) for k, v in output_headers.items()])    
    #start_response(status, str(output_headers))    

    return response([body])


# ------------------------------------------------------
# Misc. utilities
# ------------------------------------------------------

def get_base_uri(environ):
    _, filename = os.path.split(environ.get('SCRIPT_FILENAME', ''))    
    return application_uri(environ).replace('/%s' % filename, '')
    
class HTTPError(Exception):
    def __init__(self, status, headers=[], body=u''):
        self.headers = headers
        self.body = body
        super(HTTPError, self).__init__(status)

class NotFoundError(HTTPError):
    def __init__(self, headers=[], body=u''): 
        super(NotFoundError, self).__init__(HTTP_NOT_FOUND, headers, body)
                
def encode(value):
    return value.encode(ENCODING, 'replace')

def decode(value):  
    return unicode(value, ENCODING, 'replace')        
 
def redirect_to(uri, headers=None):    
    output_headers = [('Location', uri)]
    if headers: 
        output_headers.extend(headers)
    return HTTP_SEE_OTHER, output_headers, u''
        
def make_diff_link(name, revision_id):
    if revision_id > 1:
        prev_revision_id = revision_id-1
        return fill_template(u'&middot; <a rel="nofollow" title="View differences for revision $revision_id and $prev_revision_id" href="$base_uri/$name/diff/$prev_revision_id/$revision_id">Diff</a>', **locals())
    return u''
    
def make_author_link(author):
    if re_wiki_word.match(author):                         
        return fill_template(u'<a href="$base_uri/$author">$author</a>', author=author)        
    return author

def make_hash(s):      
    return hashlib.md5(encode(s)).hexdigest()

def escape_html(value):
    return cgi.escape(value, quote=True)

# def escape_url(value):
#     return urllib.quote(value)

def format_time(datetime):    
    return timeformat.format('%H:%M', 
        datetime.timetuple(), utctime=True)
        
def format_datetime(datetime, short=True):    
    return timeformat.format(short and '%b[SHORT] %d, %Y' or '%b[SHORT] %d, %Y at %H:%M', 
        datetime.timetuple(), utctime=True)

def format_iso_datetime(datetime):        
    return timeformat.format(u'%Y-%m-%dT%H:%M:%S%z', 
        datetime.timetuple(), utctime=True)

def format_log_datetime(datetime):
    return timeformat.format(u'%d/%b[SHORT]/%Y:%H:%M:%S %T', 
        datetime.timetuple(), utctime=True)

def format_http_datetime(seconds):        
    return timeformat.format(u'%a[SHORT], %d %b[SHORT] %Y %H:%M:%S %Z', 
        time.gmtime(seconds), utctime=True)

def parse_post_data(environ):
    try:
      request_body_size = int(environ.get('CONTENT_LENGTH', 0))
    except ValueError:
      request_body_size = 0
          
    request_body = environ['wsgi.input'].read(request_body_size)        
        
    d = cgi.parse_qs(request_body, keep_blank_values=True)

    def get_first(key, default=''):
        return d.get(key, [default])[0]

    return get_first


def parse_query_string(environ):

    d = cgi.parse_qs(environ.get('QUERY_STRING', ''))
          
    def get_first(key, default=''):
        return d.get(key, [default])[0]

    return get_first
    
    
def sanitize(value, escape=True, truncate=30):     
    if truncate:
        value = value[:truncate]
    
    return escape_html(value) if escape else value
    

# ------------------------------------------------------
# Cookies
# ------------------------------------------------------

def get_cookie(environ):
    return SimpleCookie(environ.get('HTTP_COOKIE', ''))

def make_cookie_headers(cookie):
    headers = []

    for c in cookie.values():
        headers.append(('Set-Cookie', c.output(header='')))

    return headers

def set_cookie(cookie, name, value, path='/', domain=None, expiration=None):

    # See http://webpython.codepoint.net/cgi_morsel
    cookie[name] = encode(value)
    cookie[name]["path"] = path    
    
    if domain:
        cookie[name]["domain"] = domain
    if expiration:
        cookie[name]["expires"] = format_http_datetime(expiration) 
        
def set_author(cookie, name):
    set_cookie(cookie, 'author', name, expiration=time.time()+60*60*24*30) # 1 month

def get_author(environ):
    cookie = get_cookie(environ)
    try:
        return sanitize(decode(cookie['author'].value))
    except KeyError:            
        pass    
    return ANONYNOUS_USER

def set_message(cookie, text):
    set_cookie(cookie, 'message', text, expiration=time.time()+60)
    
def get_message(environ, default=''):
    cookie = get_cookie(environ)
    try:
        message = decode(cookie['message'].value)
    except KeyError:
        message = default
    return message

def make_message(text):
    try: 
        klass, message = escape_html(text).split(u' ', 1)
    except ValueError:
        return text
    return u'<div class="message %s">%s</div>' % (klass.lower(), message)
        
# ------------------------------------------------------
# Templates
# ------------------------------------------------------

def fill_template_namespace(**kwargs):            
    d = kwargs.copy()
    def _(source, **kwargs):
        t = string.Template(source)    
        d.update(kwargs)        
        return t.safe_substitute(d)
    return _

def fill_template(source, **kwargs):    
    t = string.Template(source)    
    return t.safe_substitute(kwargs)

def load_disk_template(filename):            
    for template_dir in TEMPLATE_PATH:
        try:
            with codecs.open(os.path.join(template_dir, filename), encoding=ENCODING) as f:
                return f.read()
        except IOError:
            pass # Try another dir
    raise IOError('Could not find template "%s" in the following directories: %s' % (filename, ', '.join(TEMPLATE_PATH)))


nav_names = config.get('wiki', 'nav_names').split()

nav_links = u' &middot; '.join([fill_template(u'<a href="$base_uri/$name">$name</a>', name=name) for name in nav_names])
nav_links_footer = u'\n'.join([fill_template(u'<li><a href="$base_uri/$name">$name</a></li>', name=name) for name in nav_names])

# Load and fill once
meta = load_disk_template('_meta.html')
header = fill_template(load_disk_template('_header.html'), nav_links=nav_links)
footer = fill_template(load_disk_template('_footer.html'), 
            owner=make_author_link(OWNER_NAME), 
            contact=(OWNER_URI and '&nbsp;(<a href="%s">contact</a>)' % OWNER_URI or u''),
            nav_links_footer=nav_links_footer)  

def make_page(filename, filler, symbols):
    
    d = {
         # CSS class
         'page_class': symbols.get('name', '').lower(),
         # Fallback to SITE_NAME if page_title is not given
         'page_title': symbols.get('page_title', config.get('wiki', 'wiki_name'))        
    }
                           
    d.update(symbols)            
    # Assemble page _partials
    body = filler(load_disk_template(filename), **d) 

    # Resolve remaning global symbols
    return filler(body) 



# ------------------------------------------------------
# Creole processing
# ------------------------------------------------------

def load_user_macros(modulename):
    try:
        module = __import__(modulename)
        macros = dict([(macro, getattr(module, macro)) for macro in module.__all__])
    except ImportError:
        macros = None

    return macros

def load_builtin_macros():
    import macros
    return dict([(macro, getattr(macros, macro)) for macro in macros.__all__])
    
def process_content(text, macros=None, verbatim_macros=False):

    from bikini.creole.creole2html.parser import CreoleParser
    from bikini.creole.creole2html.emitter import HtmlEmitter
    
    class BikiniHtmlEmitter(HtmlEmitter): 
    
        def __init__(self, root,  macros=None, verbatim_macros=False, verbose=1, stderr=sys.stderr):
            HtmlEmitter.__init__(self, root, macros, verbose, stderr)
            # Track internal links
            self.outbound_links = []        
            self.verbatim_macros = verbatim_macros

        def macro_emit(self, node):
            if self.verbatim_macros:
                return '&lt;&lt;%s %s&gt;&gt;' % (node.macro_name, node.macro_args)
            else:
                return HtmlEmitter.macro_emit(self, node)
                
        macro_inline_emit = macro_emit
        macro_block_emit = macro_emit
    
        def link_emit(self, node):
            target = node.content
            # Check if it is a interwiki link [[WikiWord:Something]]
#            interwiki, _, rest = target.partition(':')
#             if re_wiki_word.match(target):                
#                 self.outbound_links.append(target)
            # Check if it is a [[WikiWord]]
            if re_wiki_word.match(target):                
                self.outbound_links.append(target)

            return HtmlEmitter.link_emit(self, node)
    
    parser = CreoleParser(text)
    document = parser.parse()

    return BikiniHtmlEmitter(document, macros=macros, verbatim_macros=verbatim_macros)

# ------------------------------------------------------
# Exception middleware
# ------------------------------------------------------

class ExceptionMiddleware(object):
    '''
    Sends out an exception traceback if something goes wrong.
                
    See: http://lucumr.pocoo.org/2007/5/21/getting-started-with-wsgi/
    '''
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        """Call the application can catch exceptions."""
        appiter = None
        # just call the application and send the output back
        # unchanged but catch exceptions
        try:
            appiter = self.app(environ, start_response)
            for item in appiter:
                yield item
        # if an exception occours we get the exception information
        # and prepare a traceback we can render
        except:
            e_type, e_value, tb = sys.exc_info()
            traceback = ['Traceback (most recent call last):']
            traceback += format_tb(tb)
            traceback.append('%s: %s' % (e_type.__name__, e_value))
            # we might have not a stated response by now. try
            # to start one with the status code 500 or ignore an
            # raised exception if the application already started one.
            try:
                start_response(HTTP_INTERNAL_SERVER_ERROR, [
                               ('Content-Type', 'text/plain')])
            except:
                pass
            yield '\n'.join(traceback)

        # wsgi applications might have a close function. If it exists
        # it *must* be called.
        if hasattr(self.app, 'close'):
            self.app.close()

# Install ExceptionMiddleware
dispatch_request = ExceptionMiddleware(dispatch_request)            

# ------------------------------------------------------
# Maintenance
# ------------------------------------------------------

def run_maintenance(user, hostname, rebuild_pages=False, purge_pages=False, purge_revisions=False):

    log = codecs.open(LOG_PATH, 'a', encoding=ENCODING)

    now = datetime.utcnow()  

    builtin_macros = load_builtin_macros()
    conn = storage.connect()

    counters = [0, 0, 0]

    # 1. Purge deleted pages
#@@TODO     if purge_pages:
#         pass

    # 2. Purge archived revisions
    if purge_revisions:
        with conn:
            counters[1] = storage.purge_revisions(conn, REVISION_LIFESPAN_DAYS)

    # 3. Rebuild pages
    if rebuild_pages:
        with conn:
            for name in storage.get_all_names(conn):            
                page = storage.get_page(conn, name)
                emitter = process_content(page['content'], macros=builtin_macros)    
                # Wrap everything up within a div -  
                #  this helps further processig with DOM parsers 
                content_html = u'<div>\n%s\n</div>' % emitter.emit()                
                storage.update_page(conn, name, page['title'], content_html, page['content_hash'])
                storage.update_links(conn, name, emitter.outbound_links)
                counters[2] = counters[2] + 1
        
    
    # Line roughly resembles the NCSA Common Log Format http://en.wikipedia.org/wiki/Common_Log_Format
    log.write(u'%s - %s [%s] - %s %s %s\n' % (hostname or '-', user or '-', format_log_datetime(now), counters[0], counters[1], counters[2]))
    
    log.close()
    
    return counters 

# ------------------------------------------------------
# Setup
# ------------------------------------------------------


def run_setup():

    builtin_macros = load_builtin_macros()
    
    conn = storage.connect()
    storage.setup_db(conn)
    
    for filename in os.listdir(SETUP_PATH):
        
        name, ext = os.path.splitext(filename)
        
        # Sanity check
        if not re_wiki_word.match(name):
            continue
        
        content = codecs.open(os.path.join(SETUP_PATH, filename), 'r',  encoding=ENCODING).read()
        content_html = process_content(content, macros=builtin_macros).emit() 
        
        with conn:            
            storage.add_page(conn, name, name, content_html, make_hash(content))    
            storage.add_revision(conn, name, 1, content, 'AnonymousCoward', 'localhost', message='Setup.')

            
# ------------------------------------------------------
# All set, import views
# ------------------------------------------------------

import views


